#!/bin/env python3
import math
from multiprocessing import Queue, Process
from urllib.parse import urljoin

from decimal import Decimal
import random as rnd
import time
import pygame
from pygame.locals import *

#import MultiNEAT as NEAT
import requests
import sys

"""
params = NEAT.Parameters()
params.PopulationSize = 150
params.DynamicCompatibility = True
params.AllowClones = False
params.CompatTreshold = 5.0
params.CompatTresholdModifier = 0.3
params.YoungAgeTreshold = 15
params.SpeciesMaxStagnation = 100
params.OldAgeTreshold = 35
params.MinSpecies = 3
params.MaxSpecies = 10
params.RouletteWheelSelection = True
params.RecurrentProb = 0.0
params.OverallMutationRate = 0.02
params.MutateWeightsProb = 0.90
params.WeightMutationMaxPower = 1.0
params.WeightReplacementMaxPower = 5.0
params.MutateWeightsSevereProb = 0.5
params.WeightMutationRate = 0.75
params.MaxWeight = 20
params.MutateAddNeuronProb = 0.01
params.MutateAddLinkProb = 0.02
params.MutateRemLinkProb = 0.00
params.DivisionThreshold = 0.5
params.VarianceThreshold = 0.03
params.BandThreshold = 0.3
params.InitialDepth = 3
params.MaxDepth = 4
params.IterationLevel = 1
params.Leo = True
params.GeometrySeed = True
params.LeoSeed = True
params.LeoThreshold = 0.3
params.CPPN_Bias = -3.0
params.Qtree_X = 0.0
params.Qtree_Y = 0.0
params.Width = 1.
params.Height = 1.
params.Elitism = 0.1
params.CrossoverRate = 0.5
params.MutateWeightsSevereProb = 0.01
rng = NEAT.RNG()
rng.TimeSeed()
"""

class Landing:
    server = None
    max_speed = None
    network = None
    genome = None
    status = None

    def __init__(self, url, max_speed=None, genome=None):
        self.server = url
        self.max_seed = max_speed
        self.genome = genome

    def parse_response(self, text):
        response = text.splitlines()
        if response[0] == "Zdar!":
            self.status = True
            return
        elif response[0] == "Krach!":
            self.status = False
            return
        self.id = response[0]

        self.x, self.y = response[1].split()
        self.x = float(self.x)
        self.y = float(self.y)

        self.vx, self.vy = response[2].split()
        self.vx = float(self.vx)
        self.vy = float(self.vy)

        self.fuel = float(response[3])

        self.pod_left, self.pod_right = response[4].split()
        self.pod_left = float(self.pod_left)
        self.pod_right = float(self.pod_right)

    def start(self):
        try:
            self.parse_response(requests.get(urljoin(self.server, "start")).text)
        except Exception as e:
            print("ERROR: Spatna adresa")
            print(str(e))
            sys.exit(1)

    def stav(self, up=False, left=False, right=False):
        payload = {
            "id": self.id,
            "nahoru": 1 if up else 0,
            "doleva": 1 if left else 0,
            "doprava": 1 if right else 0
        }
        self.parse_response(requests.get(urljoin(self.server, "stav"), params=payload).text)

    def simulate_server(self, time, up=False, left=False, right=False):
        if self.fuel <= 0:
            up = False
            left = False
            right = False
            self.fuel = 0.0
        if self.y <= 10.0:  # landing
            if self.get_position_to_pad == 0 and abs(self.vx) <= 5 and abs(self.vy) <= 5:
                self.status = True
            else:
                self.status = False
            self.y = 10.0
            return

        up = 1 if up else 0
        left = 1 if left else 0
        right = 1 if right else 0
        self.x = self.x + self.vx * time + 0.5 * (6 * right - 6 * left) * time**2
        self.y = self.y + self.vy * time + 0.5 * (-7 + 12 * up) * time**2
        self.vx = self.vx + (6 * right - 6 * left) * time
        self.vy = self.vy + (-7 + 12 * up) * time
        self.fuel = self.fuel - (0.5 * left + 0.5 * right + 1 * up) * time

    def simulate_start(self):
        self.id = None
        self.x = rnd.randint(10, 189) + rnd.randint(0, 100) / 100.0
        self.y = rnd.randint(90, 189) + rnd.randint(0, 100) / 100.0
        self.vx = rnd.randint(-10, 9) + rnd.randint(0, 100) / 100.0
        self.vy = rnd.randint(-10, 9) + rnd.randint(0, 100) / 100.0
        self.fuel = 20.0
        self.pod_left = rnd.randint(5, 194) + rnd.randint(0, 100) / 100.0
        self.pod_right = (self.pod_left + rnd.randint(5, 194) + rnd.randint(0, 100) / 100.0) % 195

        # vertical landing
        self.x = self.pod_left + (self.pod_right - self.pod_left) / 2
        self.vx = 0

    def get_position_to_pad(self):
        if self.pod_left + 10 <= self.x <= self.pod_right - 10:
            return 0  # rocket is above landing point
        if self.x < self.pod_left + 10:
            return (self.pod_left + 10) - self.x
        else:
            return (self.pod_right - 10) - self.x

    def gen_inputs(self):
        """NN Inputs:

        1. get_position_to_pad / 200
        2. log(y-10)
        3. Vx / 200
        4. Vy / 200
        5. activation
        """
        inputs = [
            self.get_position_to_pad() / 200,
            math.log(max(self.y - 10, 1)),
            self.vx / 200,
            self.vy / 200,
            1.0  # Activation input
        ]
        return inputs

    def get_outputs(self):
        """NN outputs:

        1. go up? n > 0.5
        2. go left? n > 0.5
        3. go right? n > 0.5
        """
        self.network.Input(self.gen_inputs())
        self.network.Activate()
        return self.network.Output()

    def interact_with_server(self, outputs):
        self.stav(outputs[0] > 0.5, outputs[1] > 0.5, outputs[2] > 0.5)

    def interact_with_fake(self, outputs, time):
        self.simulate_server(time, outputs[0] > 0.5, False, False)

    def get_fittness(self):
        """Calculate NN fittness."""
        fittness = 100
        fittness -= abs(self.get_position_to_pad())  # how far did we landed to pad
        if fittness == 100:
            fittness += 100  # we "landed" on spot
            if self.status == False:
                fittness -= abs(self.vy)  # the lower the speed we land the better
            else:
                fittness += 1000
                fittness += self.fuel * 2

        return float(fittness)
"""
    def evaluate(self):
        self.network = NEAT.NeuralNetwork()
        self.genome.BuildPhenotype(self.network)

        self.start()
        clock = pygame.time.Clock()
        while self.status == None:
            self.interact_with_server(self.get_outputs())
            clock.tick(3)

        return self.get_fittness()

    def evaluate_fake(self):
        self.network = NEAT.NeuralNetwork()
        self.genome.BuildPhenotype(self.network)

        self.simulate_start()
        while self.status == None:
            self.interact_with_fake(self.get_outputs(), 1 / 3.0)

        return self.get_fittness()
"""
"""

def epoch(pop, best_genome_ever=(Decimal("-Inf"), None), n=10):
    best_genome_ever = (Decimal("-Inf"), None)
    for generation in range(n):
        print("Generation:", generation)

        now = time.time()
        genome_list = []
        for s in pop.Species:
            for i in s.Individuals:
                genome_list.append(i)

        print('All individuals:', len(genome_list))

        for genome in genome_list:
            total_fitness = 0
            for trial in range(20):  # Each neural network have 20 trials
                landing = Landing(None, genome=genome)
                total_fitness += landing.evaluate_fake()
            genome.SetFitness(total_fitness / 20)

        best = max([(x.GetLeader().GetFitness(), x.GetLeader())
                    for x in pop.Species], key=lambda x: x[0])
        if best[0] > best_genome_ever[0]:
            best_genome_ever = best
        print('Best fitness:', best[0], 'Species:', len(pop.Species))

        print("Evaluation took", time.time() - now, "seconds.")
        print("Reproducing..")
        now = time.time()
        pop.Epoch()
        print("Reproduction took", time.time() - now, "seconds.")

    return best_genome_ever, pop


def learn():
    g = NEAT.Genome(0, 5, 0, 3, False, NEAT.ActivationFunction.TANH,
                    NEAT.ActivationFunction.UNSIGNED_SIGMOID, 0, params)
    pop = NEAT.Population(g, params, True, 1.0, rnd.randint(0, 1000))

    return pop
"""
WHITE = (255, 255, 255)
BLACK = (0,   0,   0)
BRIGHTRED = (255,   0,   0)
RED = (155,   0,   0)
BRIGHTGREEN = (0, 255,   0)
GREEN = (0, 155,   0)
BRIGHTBLUE = (0,   0, 255)
BLUE = (0,   0, 155)
BRIGHTYELLOW = (255, 255,   0)
YELLOW = (155, 155,   0)
DARKGRAY = (40,  40,  40)
bgColor = WHITE


def terminate():
    pygame.quit()
    sys.exit()


def checkForQuit():
    for event in pygame.event.get(QUIT):
        terminate()
    for event in pygame.event.get(KEYUP):
        if event.key == K_ESCAPE:
            terminate()
        pygame.event.post(event)


def UI(url):
    pygame.init()
    FPSCLOCK = pygame.time.Clock()
    DISPLAYSURF = pygame.display.set_mode((600, 600))
    pygame.display.set_caption('Space')

    spaceship = pygame.Rect(10, 10, 10, 10)

    l = Landing(url)
    l.start()
    bottom = pygame.Rect(l.pod_left, 595, l.pod_right - l.pod_left, 10)
    up = False
    right = False
    left = False
    BASICFONT = pygame.font.Font('freesansbold.ttf', 60)
    infoOverSurf = BASICFONT.render('GAMEOVER', 1, RED)
    infoOverRect = infoOverSurf.get_rect()
    infoOverRect.topleft = (10, 300)
    infoGoodSurf = BASICFONT.render('WIN', 1, RED)
    infoGoodRect = infoGoodSurf.get_rect()
    infoGoodRect.topleft = (10, 300)

    while True:
        for event in pygame.event.get():  # event handling loop
            if event.type == KEYDOWN:
                if event.key == K_UP:
                    up = True
                elif event.key == K_LEFT:
                    left = True
                elif event.key == K_RIGHT:
                    right = True
            elif event.type == KEYUP:
                if event.key == K_UP:
                    up = False
                elif event.key == K_LEFT:
                    left = False
                elif event.key == K_RIGHT:
                    right = False

        DISPLAYSURF.fill(bgColor)
        if l.status == True:
            DISPLAYSURF.blit(infoGoodSurf, infoGoodRect)
        elif l.status == False:
            DISPLAYSURF.blit(infoOverSurf, infoOverRect)
        l.stav(up, left, right)
        spaceship.x = l.x
        spaceship.y = 595 - l.y
        pygame.draw.rect(DISPLAYSURF, BLACK, spaceship)
        pygame.draw.rect(DISPLAYSURF, RED, bottom)
        pygame.display.update()
        FPSCLOCK.tick(15)
        checkForQuit()

if __name__ == "__main__":
    print("Ovladani:")
    print("SIPKY nahoru, doleva, doprava")
    UI(input("Zadejte adresu: "))
